@@ -7,7 +7,9 @@
 #include <gfarm/gfarm.h>
 
 #include "thrsubr.h"
+#include "nanosec.h"
 
+#include "timespec.h"
 #include "context.h"
 #include "config.h"
 #include "gfm_proto.h"
@@ -201,6 +203,7 @@ dfc_workq_host_remove(struct dfc_workq *q, struct host *host, const char *diag)
 
 	gfarm_mutex_lock(&q->mutex, diag, "lock");
 
+	/* dfc_allq.mutex prevents dfc from being freed */
 	for (dfc = q->q.workq_next; dfc != &q->q; dfc = next) {
 		next = dfc->workq_next;
 
@@ -307,7 +310,7 @@ handle_removal_result(struct dead_file_copy *dfc)
 	if (dfc->result == GFARM_ERR_NO_ERROR ||
 	    dfc->result == GFARM_ERR_NO_SUCH_FILE_OR_DIRECTORY) {
 		dead_file_copy_free(dfc); /* sleeps to wait for dbq.mutex */
-	} else if (host_is_up(dfc->host)) {
+	} else if (host_is_file_removable(dfc->host)) {
 		/* unexpected error, try again later to avoid busy loop */
 		gflog_notice(GFARM_MSG_1002223,
 		    "waiting removal of (%lld, %lld, %s): %s",
@@ -328,6 +331,15 @@ handle_removal_result(struct dead_file_copy *dfc)
 static void *
 removal_finalizer(void *arg)
 {
+	gfarm_error_t e;
+	static const char diag[] = "removal_finalizer";
+
+	e = gfarm_pthread_set_priority_minimum(diag);
+	if (e != GFARM_ERR_NO_ERROR)
+		gflog_info(GFARM_MSG_1005035,
+		    "%s: set_priority_minimum: %s",
+		    diag, gfarm_error_string(e));
+
 	for (;;) {
 		handle_removal_result(removal_finishedq_dequeue());
 	}
@@ -419,6 +431,7 @@ dead_file_copy_host_busyq_scan(void)
 
 	now = time(NULL);
 
+	/* dfc_allq.mutex prevents dfc from being freed */
 	for (dfc = host_busyq.q.workq_next; dfc != &host_busyq.q; dfc = next) {
 		next = dfc->workq_next;
 
@@ -440,7 +453,8 @@ dead_file_copy_host_busyq_scan(void)
 		} else if (!busy) {
 			dfc_workq_remove(dfc);
 
-			if (dead_file_copy_is_removable(dfc)) {
+			if (dead_file_copy_is_removable(dfc) &&
+			    host_is_file_removable(dfc->host)) {
 				removal_pendingq_enqueue(dfc);
 			} else {
 				/* leave it as dfcstate_deferred */
@@ -541,6 +555,15 @@ dead_file_copy_lookup(gfarm_ino_t inum, gfarm_uint64_t igen, struct host *host,
 	return (found ? dfc : NULL);
 }
 
+static struct gfarm_timespec remove_scan_log_time = { 0, 0 };
+static struct gfarm_timespec remove_scan_max_time = { 0, 0 };
+static struct gfarm_timespec remove_scan_total_time = { 0, 0 };
+static unsigned long long remove_scan_max_process_count = 0;
+static unsigned long long remove_scan_max_scan_count = 0;
+static unsigned long long remove_scan_total_process_count = 0;
+static unsigned long long remove_scan_total_scan_count = 0;
+static unsigned long long remove_scan_total_calls = 0;
+
 /*
  * PREREQUISITE: giant_lock
  * LOCKS: dfc_allq.mutex, removal_pendingq.mutex
@@ -550,7 +573,7 @@ dead_file_copy_lookup(gfarm_ino_t inum, gfarm_uint64_t igen, struct host *host,
  *
  * dead_file_copy_is_removable() needs giant_lock, and may sleep.
  */
-static void
+static long long
 dead_file_copy_scan_deferred(gfarm_ino_t inum, struct host *host,
 	int (*filter)(struct dead_file_copy *, gfarm_ino_t, struct host *),
 	const char *diag)
@@ -559,6 +582,11 @@ dead_file_copy_scan_deferred(gfarm_ino_t inum, struct host *host,
 	time_t now = time(NULL);
 	int busy, already_removed;
 	enum dead_file_copy_state state;
+	struct timespec t1, t2;
+	struct gfarm_timespec gt1, gt2, elapsed, next_log_time;
+	unsigned long long process_count = 0, scan_count = 0;
+
+	gfarm_gettime(&t1);
 
 	gfarm_mutex_lock(&dfc_allq.mutex, diag, "lock");
 
@@ -566,6 +594,7 @@ dead_file_copy_scan_deferred(gfarm_ino_t inum, struct host *host,
 	for (dfc = dfc_allq.q.allq_next; dfc != &dfc_allq.q;
 	     dfc = dfc->allq_next) {
 
+		++scan_count;
 		gfarm_mutex_lock(&dfc->mutex, diag, "dfc state");
 		state = dfc->state;
 		gfarm_mutex_unlock(&dfc->mutex, diag, "dfc state");
@@ -589,11 +618,13 @@ dead_file_copy_scan_deferred(gfarm_ino_t inum, struct host *host,
 		gfarm_mutex_unlock(&dfc_allq.mutex,
 		    diag, "unlock before sleeping");
 
+		++process_count;
 		if (already_removed) {
 			removal_finishedq_enqueue(dfc, GFARM_ERR_NO_ERROR);
 		} else {
 			busy = host_check_busy(dfc->host, now);
-			if (!busy && dead_file_copy_is_removable(dfc))
+			if (!busy && dead_file_copy_is_removable(dfc) &&
+			    host_is_file_removable(dfc->host))
 				removal_pendingq_enqueue(dfc);
 		}
 
@@ -602,6 +633,64 @@ dead_file_copy_scan_deferred(gfarm_ino_t inum, struct host *host,
 	}
 
 	gfarm_mutex_unlock(&dfc_allq.mutex, diag, "unlock");
+
+	/* the static variables below are protected by giant_lock */
+
+	gfarm_gettime(&t2);
+	gt1.tv_sec = t1.tv_sec; gt1.tv_nsec = t1.tv_nsec;
+	gt2.tv_sec = t2.tv_sec; gt2.tv_nsec = t2.tv_nsec;
+
+	elapsed = gt2;
+	gfarm_timespec_sub(&elapsed, &gt1);
+
+	if (gfarm_timespec_cmp(&remove_scan_max_time, &elapsed) < 0) {
+		remove_scan_max_time = elapsed;
+		remove_scan_max_process_count = process_count;
+		remove_scan_max_scan_count = scan_count;
+	}
+
+	gfarm_timespec_add(&remove_scan_total_time, &elapsed);
+	remove_scan_total_process_count += process_count;
+	remove_scan_total_scan_count += scan_count;
+	remove_scan_total_calls++;
+
+	/*
+	 * calculate next_log_time here to immediately reflect the change to
+	 * gfarm_metadb_remove_scan_log_interval by "gfstatus -Mm"
+	 */
+	next_log_time = remove_scan_log_time;
+	next_log_time.tv_sec += gfarm_metadb_remove_scan_log_interval;
+	if (gfarm_timespec_cmp(&gt2, &next_log_time) >= 0) {
+		gflog_info(GFARM_MSG_1005036,
+		    "dead_file_copy scan: "
+		    "max: %g seconds (%llu scans %llu processed), "
+		    "%llu times average: %g seconds (%g scans %g processed)",
+		    remove_scan_max_time.tv_sec +
+		    (double)remove_scan_max_time.tv_nsec
+		    / GFARM_SECOND_BY_NANOSEC,
+		    remove_scan_max_scan_count,
+		    remove_scan_max_process_count,
+		    remove_scan_total_calls,
+		    (remove_scan_total_time.tv_sec +
+		     (double)remove_scan_total_time.tv_nsec
+		     / GFARM_SECOND_BY_NANOSEC) / remove_scan_total_calls,
+		    (double)remove_scan_total_scan_count
+		    / remove_scan_total_calls,
+		    (double)remove_scan_total_process_count
+		    / remove_scan_total_calls);
+		remove_scan_log_time = gt2;
+		remove_scan_max_time.tv_sec = 0;
+		remove_scan_max_time.tv_nsec = 0;
+		remove_scan_max_process_count = 0;
+		remove_scan_max_scan_count = 0;
+		remove_scan_total_time.tv_sec = 0;
+		remove_scan_total_time.tv_nsec = 0;
+		remove_scan_total_process_count = 0;
+		remove_scan_total_scan_count = 0;
+		remove_scan_total_calls = 0;
+	}
+	return ((long long)(t2.tv_sec - t1.tv_sec) * GFARM_SECOND_BY_NANOSEC
+	    + (t2.tv_nsec - t1.tv_nsec));
 }
 
 static int
@@ -618,12 +707,13 @@ transparent_filter(struct dead_file_copy *dfc,
  *	but dfc_allq.mutex and removal_pendingq.mutex
  *	won't be blocked while sleeping.
  */
-void
-dead_file_copy_scan_deferred_all(void)
+static long long
+dead_file_copy_scan_deferred_all_run(void)
 {
-	static const char diag[] = "dead_file_copy_scan_deferred_all";
+	static const char diag[] = "dead_file_copy_scan_deferred_all_run";
 
-	dead_file_copy_scan_deferred(0, NULL, transparent_filter, diag);
+	return (dead_file_copy_scan_deferred(
+	    0, NULL, transparent_filter, diag));
 
 	/* leave the host_busyq as is, because it will be handled shortly. */
 }
@@ -657,10 +747,10 @@ dead_file_copy_host_becomes_down(struct host *host)
  *	but dfc_allq.mutex and removal_pendingq.mutex
  *	won't be blocked while sleeping.
  */
-void
-dead_file_copy_host_becomes_up(struct host *host)
+static long long
+dead_file_copy_host_becomes_up_run(struct host *host)
 {
-	static const char diag[] = "dead_file_copy_host_becomes_up";
+	static const char diag[] = "dead_file_copy_host_becomes_up_run";
 
 	/*
 	 * in this case, we have to check dead_file_copy data for *all* hosts,
@@ -674,7 +764,8 @@ dead_file_copy_host_becomes_up(struct host *host)
 	 *	have pending dead_file_copy entries which processing is
 	 *	deferred because only owner of the newest replica is down.
 	 */
-	dead_file_copy_scan_deferred(0, NULL, transparent_filter, diag);
+	return (dead_file_copy_scan_deferred(
+	    0, NULL, transparent_filter, diag));
 
 	/* leave the host_busyq as is, because it will be handled shortly. */
 }
@@ -756,12 +847,13 @@ inode_and_host_filter(struct dead_file_copy *dfc,
  * this assumes that the number of dead file copies is small enough,
  * otherwise this is too slow.
  */
-void
-dead_file_copy_replica_status_changed(gfarm_ino_t inum, struct host *host)
+static long long
+dead_file_copy_replica_status_changed_run(gfarm_ino_t inum, struct host *host)
 {
-	static const char diag[] = "dead_file_copy_replica_status_changed";
+	static const char diag[] = "dead_file_copy_replica_status_changed_run";
 
-	dead_file_copy_scan_deferred(inum, host, inode_and_host_filter, diag);
+	return (dead_file_copy_scan_deferred(
+	    inum, host, inode_and_host_filter, diag));
 }
 
 static int
@@ -782,14 +874,202 @@ inode_filter(struct dead_file_copy *dfc,
  * this assumes that the number of dead file copies is small enough,
  * otherwise this is too slow.
  */
+static long long
+dead_file_copy_inode_status_changed_run(gfarm_ino_t inum)
+{
+	static const char diag[] = "dead_file_copy_inode_status_changed_run";
+
+	return (dead_file_copy_scan_deferred(
+	    inum, NULL, inode_filter, diag));
+}
+
+static struct {
+	pthread_mutex_t mutex;
+	pthread_cond_t requested;
+	int need_to_run;
+	gfarm_ino_t inum;
+	struct host *host;
+} dead_file_copy_scan = {
+	PTHREAD_MUTEX_INITIALIZER,
+	PTHREAD_COND_INITIALIZER,
+	0,
+	0,
+	NULL
+};
+
+static const char dead_file_copy_scan_diag[] = "dead_file_copy_scan";
+static const char dead_file_copy_scan_requested_diag[] =
+	"dead_file_copy_scan_requested";
+
+void
+dead_file_copy_scan_deferred_all(void)
+{
+	static const char diag[] = "dead_file_copy_scan_deferred_all";
+
+	gfarm_mutex_lock(&dead_file_copy_scan.mutex,
+	    diag, dead_file_copy_scan_diag);
+
+	/* scan all */
+	dead_file_copy_scan.inum = 0;
+	dead_file_copy_scan.host = NULL;
+
+	dead_file_copy_scan.need_to_run = 1;
+	gfarm_cond_signal(
+	    &dead_file_copy_scan.requested, diag,
+	    dead_file_copy_scan_requested_diag);
+
+	gfarm_mutex_unlock(&dead_file_copy_scan.mutex,
+	    diag, dead_file_copy_scan_diag);
+}
+
+void
+dead_file_copy_host_becomes_up(struct host *host)
+{
+	static const char diag[] = "dead_file_copy_host_becomes_up";
+
+	gfarm_mutex_lock(&dead_file_copy_scan.mutex,
+	    diag, dead_file_copy_scan_diag);
+
+	if (!dead_file_copy_scan.need_to_run ||
+	    dead_file_copy_scan.host == host) {
+		/* only this host */
+		dead_file_copy_scan.inum = 0;
+		dead_file_copy_scan.host = host;
+	} else { /* otherwise scan all */
+		dead_file_copy_scan.inum = 0;
+		dead_file_copy_scan.host = NULL;
+	}
+
+	dead_file_copy_scan.need_to_run = 1;
+	gfarm_cond_signal(
+	    &dead_file_copy_scan.requested, diag,
+	    dead_file_copy_scan_requested_diag);
+
+	gfarm_mutex_unlock(&dead_file_copy_scan.mutex,
+	    diag, dead_file_copy_scan_diag);
+}
+
+void
+dead_file_copy_replica_status_changed(gfarm_ino_t inum, struct host *host)
+{
+	static const char diag[] = "dead_file_copy_replica_status_changed";
+
+	gfarm_mutex_lock(&dead_file_copy_scan.mutex,
+	    diag, dead_file_copy_scan_diag);
+
+	if (!dead_file_copy_scan.need_to_run ||
+	    (dead_file_copy_scan.inum == inum &&
+	     dead_file_copy_scan.host == host)) {
+		/* only this replica */
+		dead_file_copy_scan.inum = inum;
+		dead_file_copy_scan.host = host;
+	} else if (dead_file_copy_scan.inum == inum) {
+		/* only this inode */
+		dead_file_copy_scan.host = NULL;
+	} else { /* otherwise scan all */
+		dead_file_copy_scan.inum = 0;
+		dead_file_copy_scan.host = NULL;
+	}
+
+	dead_file_copy_scan.need_to_run = 1;
+	gfarm_cond_signal(
+	    &dead_file_copy_scan.requested, diag,
+	    dead_file_copy_scan_requested_diag);
+
+	gfarm_mutex_unlock(&dead_file_copy_scan.mutex,
+	    diag, dead_file_copy_scan_diag);
+}
+
 void
 dead_file_copy_inode_status_changed(gfarm_ino_t inum)
 {
 	static const char diag[] = "dead_file_copy_inode_status_changed";
 
-	dead_file_copy_scan_deferred(inum, NULL, inode_filter, diag);
+	gfarm_mutex_lock(&dead_file_copy_scan.mutex,
+	    diag, dead_file_copy_scan_diag);
+
+	if (!dead_file_copy_scan.need_to_run ||
+	    dead_file_copy_scan.inum == inum) {
+		/* only this inode */
+		dead_file_copy_scan.inum = inum;
+		dead_file_copy_scan.host = NULL;
+	} else { /* otherwise scan all */
+		dead_file_copy_scan.inum = 0;
+		dead_file_copy_scan.host = NULL;
+	}
+
+	dead_file_copy_scan.need_to_run = 1;
+	gfarm_cond_signal(
+	    &dead_file_copy_scan.requested, diag,
+	    dead_file_copy_scan_requested_diag);
+
+	gfarm_mutex_unlock(&dead_file_copy_scan.mutex,
+	    diag, dead_file_copy_scan_diag);
 }
 
+static void *
+dead_file_copy_scanner(void *arg)
+{
+	gfarm_error_t e;
+	static const char diag[] = "dead_file_copy_scanner";
+	gfarm_ino_t inum;
+	struct host *host;
+	long long t;
+	int interval_factor;
+
+	e = gfarm_pthread_set_priority_minimum(diag);
+	if (e != GFARM_ERR_NO_ERROR)
+		gflog_info(GFARM_MSG_1005037,
+		    "%s: set_priority_minimum: %s",
+		    diag, gfarm_error_string(e));
+
+	for (;;) {
+		gfarm_mutex_lock(&dead_file_copy_scan.mutex,
+		    diag, dead_file_copy_scan_diag);
+
+		while (!dead_file_copy_scan.need_to_run)
+			gfarm_cond_wait(
+			    &dead_file_copy_scan.requested,
+			    &dead_file_copy_scan.mutex, diag,
+			    dead_file_copy_scan_requested_diag);
+		inum = dead_file_copy_scan.inum;
+		host = dead_file_copy_scan.host;
+
+		dead_file_copy_scan.need_to_run = 0;
+		dead_file_copy_scan.inum = 0;
+		dead_file_copy_scan.host = NULL;
+
+		gfarm_mutex_unlock(&dead_file_copy_scan.mutex,
+		    diag, dead_file_copy_scan_diag);
+
+		giant_lock();
+
+		if (inum != 0 && host != NULL) {
+			t = dead_file_copy_replica_status_changed_run(
+			    inum, host);
+		} else if (inum != 0) {
+			t = dead_file_copy_inode_status_changed_run(inum);
+		} else if (host != NULL) {
+			t = dead_file_copy_host_becomes_up_run(host);
+		} else {
+			t = dead_file_copy_scan_deferred_all_run();
+		}
+
+		/* giant_lock is needed for GFM_PROTO_CONFIG_SET */
+		interval_factor = gfarm_metadb_remove_scan_interval_factor;
+
+		giant_unlock();
+
+		if (t <= 0)
+			t = 1;
+		gfarm_nanosleep(t * interval_factor);
+	}
+
+	/*NOTREACHED*/
+	return (NULL);
+}
+
+
 /* used for deferred -> kept: prevent dfc from moved to removal_pendingq */
 void
 dead_file_copy_mark_kept(struct dead_file_copy *dfc)
@@ -885,6 +1165,7 @@ dead_file_copy_count_by_inode(gfarm_ino_t inum, gfarm_uint64_t igen,
 
 	gfarm_mutex_lock(&dfc_allq.mutex, diag, "lock");
 
+	/* dfc_allq.mutex prevents dfc from being freed */
 	for (dfc = dfc_allq.q.allq_next; dfc != &dfc_allq.q;
 	     dfc = dfc->allq_next) {
 
@@ -921,6 +1202,7 @@ dead_file_copy_info_by_inode(gfarm_ino_t inum, gfarm_uint64_t igen, int up_only,
 
 	gfarm_mutex_lock(&dfc_allq.mutex, diag, "lock");
 
+	/* dfc_allq.mutex prevents dfc from being freed */
 	for (dfc = dfc_allq.q.allq_next; dfc != &dfc_allq.q;
 	     dfc = dfc->allq_next) {
 		if (i >= n)
@@ -1155,7 +1437,8 @@ dead_file_copy_add_one(void *closure,
 		    "%s: dead replica(%lld, %lld, %s): no memory",
 		    diag, (long long)inum, (long long)igen, hostname);
 #if 0 /* this is unnecessary, since all hosts are down in this point */
-	} else if (dead_file_copy_is_removable(dfc)) {
+	} else if (dead_file_copy_is_removable(dfc) &&
+	    host_is_file_removable(dfc->host)) {
 		removal_pendingq_enqueue(dfc);
 #endif
 	} /* otherwise leave it as dfcstate_deferred */
@@ -1192,4 +1475,9 @@ dead_file_copy_init(int is_master)
 		    "create_detached_thread(removal_finalizer): %s",
 		    gfarm_error_string(e));
 
+	e = create_detached_thread(dead_file_copy_scanner, NULL);
+	if (e != GFARM_ERR_NO_ERROR)
+		gflog_fatal(GFARM_MSG_1005038,
+		    "create_detached_thread(dead_file_copy_scanner): %s",
+		    gfarm_error_string(e));
 }
