@@ -1,5 +1,5 @@
 /*
- * $Id: fsngroup.c 10665 2017-11-13 10:04:19Z takuya-i $
+ * $Id$
  */
 
 #include <assert.h>
@@ -83,26 +83,57 @@ fsngroup_get_tuples(gfarm_int32_t *n_tuples, struct fsngroup_tuple **tuplesp)
 	return (GFARM_ERR_NO_ERROR);
 }
 
-static int
-fsngroup_filter(struct host *h, void *closure)
-{
-	const char *fsngroup = closure;
-
-	return (strcmp(host_fsngroup(h), fsngroup) == 0);
-}
-
-gfarm_error_t
-fsngroup_get_hosts(const char *fsngroup, int *nhostsp, struct host ***hostsp)
-{
-	return (host_from_all(fsngroup_filter, (char *)fsngroup, /* UNCONST */
-	    nhostsp, hostsp));
-}
-
 /*****************************************************************************/
 /*
  * Replication scheduler:
  */
 
+static gfarm_error_t
+gfarm_repattr_parse_cached(const char *fsng,
+	gfarm_repattr_t **repsp, size_t *nrepsp)
+{
+	/* cache */
+	static char *last_fsng = NULL;
+	static gfarm_repattr_t *last_reps = NULL;
+	static size_t last_nreps = 0;
+
+	gfarm_error_t e;
+	char *fsng_tmp;
+	static gfarm_repattr_t *reps_tmp;
+	static size_t nreps_tmp;
+
+	if (last_fsng != NULL && strcmp(fsng, last_fsng) == 0) {
+		/* cache hit */
+		*repsp = last_reps;
+		*nrepsp = last_nreps;
+		return (GFARM_ERR_NO_ERROR);
+	}
+
+	e = gfarm_repattr_parse(fsng, &reps_tmp, &nreps_tmp);
+	if (e != GFARM_ERR_NO_ERROR)
+		return (e);
+
+	/* update the cache */
+
+	fsng_tmp = strdup(fsng);
+	if (fsng_tmp == NULL) {
+		gflog_error(GFARM_MSG_1005059, "no memory for '%s'", fsng);
+		return (GFARM_ERR_NO_MEMORY);
+	}
+
+	if (last_reps != NULL)
+		gfarm_repattr_free_all(last_nreps, last_reps);
+	free(last_fsng);
+
+	last_fsng = fsng_tmp;
+	last_reps = reps_tmp;
+	last_nreps = nreps_tmp;
+
+	*repsp = last_reps;
+	*nrepsp = last_nreps;
+	return (GFARM_ERR_NO_ERROR);
+}
+
 /*
  * Make us sure our having the giant_lock acquired.
  */
@@ -113,15 +144,15 @@ gfarm_error_t
 fsngroup_schedule_replication(
 	struct inode *inode, struct dirset *tdirset, const char *repattr,
 	int n_srcs, struct host **srcs,
-	int *n_existingp, struct host **existing, gfarm_time_t grace,
-	int *n_being_removedp, struct host **being_removed, const char *diag,
+	int *n_existingp, struct hostset *existing, gfarm_time_t grace,
+	int *n_being_removedp, struct hostset *being_removed, const char *diag,
 	int *total_p, int *req_ok_nump)
 {
 	gfarm_error_t e, save_e = GFARM_ERR_NO_ERROR;
 	int i, n_scope;
 	size_t nreps = 0;
 	gfarm_repattr_t *reps = NULL;
-	struct host **scope;
+	struct hostset *scope;
 
 	assert(repattr != NULL && srcs != NULL);
 
@@ -133,7 +164,7 @@ fsngroup_schedule_replication(
 
 	*total_p = 0;
 
-	e = gfarm_repattr_parse(repattr, &reps, &nreps);
+	e = gfarm_repattr_parse_cached(repattr, &reps, &nreps);
 	if (e != GFARM_ERR_NO_ERROR) {
 		gflog_error(GFARM_MSG_1004051,
 		    "%s: %s", diag, gfarm_error_string(e));
@@ -154,26 +185,26 @@ fsngroup_schedule_replication(
 
 			*total_p = *total_p + num;
 
-			e = fsngroup_get_hosts(group, &n_scope, &scope);
-			if (e != GFARM_ERR_NO_ERROR) {
+			scope = hostset_of_fsngroup_alloc(group, &n_scope);
+			if (scope == NULL) {
+				save_e = GFARM_ERR_NO_MEMORY;
 				gflog_notice(GFARM_MSG_1004053,
-				    "%s: fsngroup_get_hosts(%s): %s",
-				    diag, group, gfarm_error_string(e));
-				if (save_e == GFARM_ERR_NO_ERROR ||
-				    e == GFARM_ERR_NO_MEMORY)
-					save_e = e;
-				continue;
+				    "%s: hostset_of_fsngroup(%s): %s",
+				    diag, group, gfarm_error_string(save_e));
+				break;
 			}
 			e = inode_schedule_replication_within_scope(
 			    inode, tdirset, num, n_srcs, srcs, &next_src_index,
 			    &n_scope, scope, n_existingp, existing, grace,
 			    n_being_removedp, being_removed, diag,
 			    req_ok_nump);
-			if (e != GFARM_ERR_NO_ERROR &&
-			    (save_e == GFARM_ERR_NO_ERROR ||
-			     e == GFARM_ERR_NO_MEMORY))
+			hostset_free(scope);
+			if (e == GFARM_ERR_NO_MEMORY) {
+				save_e = e;
+				break;
+			}
+			if (save_e == GFARM_ERR_NO_ERROR)
 				save_e = e;
-			free(scope);
 		}
 	} else {
 		/* ignore hostgroups, but the total number is used. */
@@ -184,9 +215,87 @@ fsngroup_schedule_replication(
 		}
 	}
 
-	gfarm_repattr_free_all(nreps, reps);
+	/* gfarm_repattr_free_all(nreps, reps); -- shouldn't call (cached) */
 	return (save_e);
 }
+
+/* return GFARM_ERR_NO_ERROR, if there is a spare */
+gfarm_error_t
+fsngroup_has_spare_for_repattr(struct inode *inode, int current_copy_count,
+	const char *fsng, const char *repattr, int up_only)
+{
+	gfarm_error_t e;
+	int n_scope;
+	struct hostset *scope;
+	int ncopy, n_desired, total, found;
+	size_t i, nreps = 0;
+	gfarm_repattr_t *reps = NULL;
+
+	e = gfarm_repattr_parse_cached(repattr, &reps, &nreps);
+	if (e != GFARM_ERR_NO_ERROR) {
+		gflog_error(GFARM_MSG_1004025,
+		    "gfarm_repattr_parse(%s): %s",
+		    repattr, gfarm_error_string(e));
+		return (e); /* report shortage, for safety */
+	}
+
+	n_desired = 0;
+	total = 0;
+	found = 0;
+	for (i = 0; i < nreps; i++) {
+		int num = gfarm_repattr_amount(reps[i]);
+
+		total += num;
+		if (!found &&
+		    strcmp(gfarm_repattr_group(reps[i]), fsng) == 0) {
+			n_desired = num;
+			found = 1;
+		}
+	}
+	/* gfarm_repattr_free_all(nreps, reps); -- shouldn't call (cached) */
+
+	if (current_copy_count <= total)
+		return (GFARM_ERR_INSUFFICIENT_NUMBER_OF_FILE_REPLICAS);
+
+	if (!gfarm_replicainfo_enabled) {
+		/*
+		 * ignore 'n_desired' for the hostgroup
+		 * (use 'total' only)
+		 */
+		return (GFARM_ERR_NO_ERROR); /* has spare */
+	}
+
+	/* no desired number for the host */
+	if (n_desired <= 0)
+		return (GFARM_ERR_NO_ERROR); /* has spare */
+
+	scope = hostset_of_fsngroup_alloc(fsng, &n_scope);
+	if (scope == NULL) {
+		e = GFARM_ERR_NO_MEMORY;
+		gflog_debug(GFARM_MSG_1004026,
+		    "fsngroup_get_hosts(%s): %s",
+		    fsng, gfarm_error_string(e));
+		return (e); /* report shortage, for safety */
+	}
+	if (n_scope == 0) { /* unexpected */
+		hostset_free(scope);
+		gflog_error(GFARM_MSG_1005060,
+		    "no host in fsngroup %s: unexpected", fsng);
+
+		/* report shortage, for safety */
+		return (GFARM_ERR_INTERNAL_ERROR);
+	}
+
+	ncopy = inode_count_replicas_within_scope(
+	    inode, 1, up_only, 0, scope);
+
+	hostset_free(scope);
+
+	if (ncopy > n_desired)
+		return (GFARM_ERR_NO_ERROR); /* has spare */
+	return (GFARM_ERR_INSUFFICIENT_NUMBER_OF_FILE_REPLICAS); /* shortage */
+}
+
 /*****************************************************************************/
 /*
  * Server side RPC stubs:
