@@ -191,7 +191,15 @@ static const char *journal_operation_names[] = {
 	"MDHOST_REMOVE",
 
 	"FSNGROUP_MODIFY",
+
 	"NOP",
+
+	"QUOTA_DIRSET_ADD",
+	"QUOTA_DIRSET_MODIFY",
+	"QUOTA_DIRSET_REMOVE",
+
+	"QUOTA_DIR_ADD",
+	"QUOTA_DIR_REMOVE",
 };
 
 struct journal_file_writer *
@@ -383,7 +391,7 @@ journal_file_reader_commit_pos(struct journal_file_reader *reader)
 	else if (pos == jf->tail && reader->committed_lap == jf->writer.lap)
 		reader->committed_pos = pos;
 	else {
-		reader->committed_pos = 
+		reader->committed_pos =
 		    pos - jf->tail + JOURNAL_FILE_HEADER_SIZE;
 		reader->committed_lap++;
 	}
@@ -726,7 +734,7 @@ journal_file_reader_writer_wait(struct journal_file_reader *reader,
 		if (wpos == rpos) {
 			/*
 			 * Case 1: wpos == rpos
-			 *   The writer must wait for the reader if the 
+			 *   The writer must wait for the reader if the
 			 *   reader is a lap behind.
 			 */
 			needed = (wlap > rlap);
@@ -921,7 +929,7 @@ journal_fread_file_header(FILE *file)
 
 static gfarm_error_t
 journal_find_rw_pos0(FILE *file, int has_writer, size_t file_size,
-	gfarm_uint64_t db_seqnum, off_t *rposp, gfarm_uint64_t *rlapp, 
+	gfarm_uint64_t db_seqnum, off_t *rposp, gfarm_uint64_t *rlapp,
 	off_t *wposp, gfarm_uint64_t wlap, off_t *tailp,
 	gfarm_uint64_t *max_seqnump)
 {
@@ -987,7 +995,7 @@ journal_find_rw_pos0(FILE *file, int has_writer, size_t file_size,
 			    db_seqnum <= seqnum) {
 				max_seqnum = seqnum;
 				max_seqnum_next_pos = next_pos;
-			} 
+			}
 			if (db_seqnum < begin_seqnum &&
 			    min_seqnum > begin_seqnum) {
 				min_seqnum = begin_seqnum;
@@ -1017,9 +1025,6 @@ journal_find_rw_pos0(FILE *file, int has_writer, size_t file_size,
 	/*
 	 * Check 'max_seqnum' and 'min_seqnum'.
 	 *
-	 * - '!has_writer' means that this function is called for creating
-	 *   a reader which is used for sending records to a slave gfmd.
-	 *
 	 * - 'db_seqnum != GFARM_METADB_SERVER_SEQNUM_INVALID' means that
 	 *   gfmd is about to start.  In case of gfjournal command,
 	 *   'db_seqnum' is always 0.
@@ -1040,7 +1045,7 @@ journal_find_rw_pos0(FILE *file, int has_writer, size_t file_size,
 	 * - 'max_seqnum < db_seqnum' means that database, the journal
 	 *   file or both are corrupted.
 	 */
-	if (!has_writer && db_seqnum != max_seqnum &&
+	if (db_seqnum != max_seqnum &&
 	    db_seqnum != GFARM_METADB_SERVER_SEQNUM_INVALID &&
 	    max_seqnum != GFARM_METADB_SERVER_SEQNUM_INVALID &&
 	    (min_seqnum == GFARM_UINT64_MAX || db_seqnum + 1 < min_seqnum ||
@@ -1131,6 +1136,14 @@ journal_close_fd_op(void *cookie, int fd)
 	    GFARM_ERR_NO_ERROR);
 }
 
+static gfarm_error_t
+journal_shutdown_fd_op(void *cookie, int fd)
+{
+	gflog_error(GFARM_MSG_1005095,
+	    "journal_shutdown_fd_op: this function shouldn't be called");
+	return (GFARM_ERR_NO_ERROR);
+}
+
 static gfarm_error_t
 journal_export_credential_fd_op(void *cookie)
 {
@@ -1229,7 +1242,7 @@ journal_write_file_header(int fd)
 	gfarm_uint32_t version = htonl(GFARM_JOURNAL_VERSION);
 	int sync_result;
  	off_t pos = 0;
- 
+
 	GFARM_MALLOC_ARRAY(header, JOURNAL_FILE_HEADER_SIZE);
 	if (header == NULL) {
 		e = GFARM_ERR_NO_MEMORY;
@@ -1238,7 +1251,7 @@ journal_write_file_header(int fd)
 			gfarm_error_string(e));
 		goto error;
 	}
- 
+
 	memset(header, 0, JOURNAL_FILE_HEADER_SIZE);
 	memcpy(header, GFARM_JOURNAL_FILE_MAGIC, GFARM_JOURNAL_MAGIC_SIZE);
 	memcpy(header + GFARM_JOURNAL_MAGIC_SIZE, (void *) &version,
@@ -1273,6 +1286,7 @@ error:
 
 static struct gfp_iobuffer_ops journal_iobuffer_ops = {
 	journal_close_fd_op,
+	journal_shutdown_fd_op,
 	journal_export_credential_fd_op,
 	journal_delete_credential_fd_op,
 	journal_env_for_credential_fd_op,
@@ -1283,7 +1297,7 @@ static struct gfp_iobuffer_ops journal_iobuffer_ops = {
 
 static gfarm_error_t
 journal_file_reader_init(struct journal_file *jf, int fd,
-	off_t pos, gfarm_uint64_t lap, int block_writer, 
+	off_t pos, gfarm_uint64_t lap, int block_writer,
 	struct journal_file_reader *reader)
 {
 	gfarm_error_t e;
